"""Logic Factories.

Adapted from the implementation for `mdformat-gfm-alerts`:
<https://github.com/KyleKing/mdformat-gfm-alerts/blob/a6e71db4da3a421320e75b8a9e3e3cb2dca429d7/mdformat_gfm_alerts/factories/_gfm_blockquote_factories.py>

"""

from __future__ import annotations

import re
from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Callable, NamedTuple

from markdown_it import MarkdownIt
from markdown_it.rules_block import StateBlock
from mdit_py_plugins.utils import is_code_block

if TYPE_CHECKING:
    from markdown_it.token import Token


# FYI: copied from mdformat_admon.factories
@contextmanager
def new_token(state: StateBlock, name: str, kind: str) -> Generator[Token, None, None]:
    """Creates scoped token."""
    yield state.push(f"{name}_open", kind, 1)
    state.push(f"{name}_close", kind, -1)


# FYI: Adapted from mdformat_admon.factories
class CalloutState(NamedTuple):
    """Frozen state."""

    parentType: str
    lineMax: int


class CalloutData(NamedTuple):
    """CalloutData data for rendering."""

    old_state: CalloutState
    meta_text: str
    folded: bool
    custom_title: str
    next_line: int


def parse_possible_blockquote_admon_factory(
    prefix: str,
    patterns: set[str],
) -> Callable[[StateBlock, int, int, bool], CalloutData | bool]:
    """Generate the parser function.

    Accepts set of strings that will be compiled into regular expressions.
    They must have a capture group `title` and optional group `folded`.

    """

    def parse_possible_blockquote_admon(
        state: StateBlock,
        start_line: int,
        end_line: int,
        silent: bool,
    ) -> CalloutData | bool:
        if is_code_block(state, start_line):
            return False

        start = state.bMarks[start_line] + state.tShift[start_line]

        # Exit if no match for any pattern
        text = state.src[start:]
        regexes = [
            re.compile(rf"{pat}(?P<custom_title>(?: |<br>)[^\n]+)?", re.IGNORECASE)
            for pat in patterns
        ]
        match = next((_m for rx in regexes if (_m := rx.match(text))), None)
        if not match:
            return False

        # Since start is found, we can report success here in validation mode
        if silent:
            return True

        old_state = CalloutState(
            parentType=state.parentType,
            lineMax=state.lineMax,
        )
        state.parentType = prefix

        return CalloutData(
            old_state=old_state,
            meta_text=match["title"],
            folded=bool(match["folded"]),
            custom_title=match["custom_title"] or "",
            next_line=end_line,
        )

    return parse_possible_blockquote_admon


def obsidian_callout_plugin_factory(
    prefix: str,
    logic: Callable[[StateBlock, int, int, bool], bool],
) -> Callable[[MarkdownIt], None]:
    """Generate the plugin function."""

    def obsidian_callout_plugin(md: MarkdownIt) -> None:
        md.block.ruler.before("blockquote", prefix, logic)

    return obsidian_callout_plugin



"""Obsidian Callouts."""

from __future__ import annotations

from markdown_it.rules_block import StateBlock

from ..factories import (
    CalloutData,
    new_token,
    obsidian_callout_plugin_factory,
    parse_possible_blockquote_admon_factory,
)

OBSIDIAN_CALLOUT_PREFIX = "obsidian_callout"
"""Prefix used to differentiate the parsed output."""

INLINE_SEP = "\n\n"
"""Optional separator to differentiate the title and ineline content (if present)."""

PATTERNS = {
    # Note '> ' prefix is removed when parsing
    r"^\*\*(?P<title>Note|Warning)\*\*",
    # FYI: Unlike GitHub Alerts, Obsidian doesn't constrain the possible types
    r"^\\?\[!(?P<title>[^\]]+)\\?\](?P<folded>-?)",
}
"""Patterns specific to GitHub Alerts."""


def format_obsidian_callout_markup(
    state: StateBlock,
    start_line: int,
    admonition: CalloutData,
) -> None:
    """Format markup."""
    tag = admonition.meta_text.upper()
    fold = "-" if admonition.folded else ""
    custom_title = admonition.custom_title
    title_line = f"[!{tag}]{fold}{INLINE_SEP}{custom_title}"

    with new_token(state, OBSIDIAN_CALLOUT_PREFIX, "div") as token:
        token.attrs = {
            "data-callout-metadata": "",
            "data-callout-fold": "",
            "data-callout": admonition.meta_text.lower(),
            "class": "callout",
        }
        if admonition.folded:
            token.attrs["data-callout-fold"] = "-"
            token.attrs["class"] = "callout is-collapsible is-collapsed"
        token.block = True
        token.map = [start_line, admonition.next_line]
        token.markup = title_line
        with new_token(state, f"{OBSIDIAN_CALLOUT_PREFIX}_title", "div") as tkn_title:
            tkn_title.attrs = {"class": "callout-title"}

            title_inner = f"{OBSIDIAN_CALLOUT_PREFIX}_title_inner"
            with new_token(state, title_inner, "div") as tkn_title_inner:
                tkn_title_inner.attrs = {"class": "callout-title-inner"}

                tkn_title_txt = state.push("inline", "", 0)
                tkn_title_txt.content = admonition.custom_title.strip()
            if admonition.folded:
                collapsed = f"{OBSIDIAN_CALLOUT_PREFIX}_collapsed"
                with new_token(state, collapsed, "div") as tkn_collapsed:
                    tkn_collapsed.attrs = {"class": "callout-fold is-collapsed"}

        content = f"{OBSIDIAN_CALLOUT_PREFIX}_content"
        with new_token(state, content, "div") as tkn_content:
            tkn_content.attrs = {"class": "callout-content"}
            if admonition.folded:
                tkn_content.attrs["style"] = "display: none;"

            state.md.block.tokenize(state, start_line + 1, admonition.next_line)

    # FIXME: this isn't actually replacing the block quote outer div?
    #
    # Render as a div for accessibility rather than block quote
    #   Which is because '>' is being misused (https://github.com/orgs/community/discussions/16925#discussioncomment-8729846)
    state.parentType = "div"  # admonition.old_state.parentType
    state.lineMax = admonition.old_state.lineMax
    state.line = admonition.next_line


def alert_logic(
    state: StateBlock,
    startLine: int,
    endLine: int,
    silent: bool,
) -> bool:
    """Parse GitHub Alerts."""
    parser_func = parse_possible_blockquote_admon_factory(
        OBSIDIAN_CALLOUT_PREFIX,
        PATTERNS,
    )
    result = parser_func(state, startLine, endLine, silent)
    if isinstance(result, CalloutData):
        format_obsidian_callout_markup(state, startLine, admonition=result)
        return True
    return result


obsidian_callout_plugin = obsidian_callout_plugin_factory(
    OBSIDIAN_CALLOUT_PREFIX,
    alert_logic,
)



"""Public Extension."""

from __future__ import annotations

from typing import Mapping

from markdown_it import MarkdownIt
from mdformat.renderer import RenderContext, RenderTreeNode
from mdformat.renderer.typing import Render

from .mdit_plugins import INLINE_SEP, OBSIDIAN_CALLOUT_PREFIX, obsidian_callout_plugin


def update_mdit(mdit: MarkdownIt) -> None:
    """Update the parser to identify Alerts."""
    mdit.use(obsidian_callout_plugin)


def _render_obsidian_callout(node: RenderTreeNode, context: RenderContext) -> str:
    """Render a `RenderTreeNode`."""
    title_line = node.markup.replace(INLINE_SEP, "")
    elements = [render for child in node.children if (render := child.render(context))]
    # Do not separate the title line from the first row
    return "\n".join([title_line, "\n\n".join(elements)]).rstrip()


def _no_render(
    node: RenderTreeNode,  # noqa: ARG001
    context: RenderContext,  # noqa: ARG001
) -> str:
    """Skip rendering when handled separately."""
    return ""


def _recursive_render(
    node: RenderTreeNode,
    context: RenderContext,
) -> str:
    elements = [render for child in node.children if (render := child.render(context))]
    # Do not separate the title line from the first row
    return "\n\n".join(elements).rstrip()


# A mapping from syntax tree node type to a function that renders it.
# This can be used to overwrite renderer functions of existing syntax
# or add support for new syntax.
RENDERERS: Mapping[str, Render] = {
    OBSIDIAN_CALLOUT_PREFIX: _render_obsidian_callout,
    f"{OBSIDIAN_CALLOUT_PREFIX}_title": _no_render,
    f"{OBSIDIAN_CALLOUT_PREFIX}_title_inner": _no_render,
    f"{OBSIDIAN_CALLOUT_PREFIX}_collapsed": _no_render,
    # FIXME: can I add divs without introducing new blocks?
    f"{OBSIDIAN_CALLOUT_PREFIX}_content": _recursive_render,
}
